import { Component, Vue, Prop, Watch } from 'vue-property-decorator';

const VECTOR_X = 0;
const VECTOR_Y = 1;
const VECTOR_DRAG = 2;
const VECTOR_SIZE = 3;
const VECTOR_COLOR = 4;
const VECTOR_OPACITY = 5;

export interface CanvasModel {
  width: number;
  height: number;
  vectors: string[][];
  imgUrl?: string;
}

export const defaultCanvasModel: CanvasModel = {
  width: 500,
  height: 800,
  vectors: [],
  imgUrl: '',
};

@Component
export default class CanvasInputComponent extends Vue {
  @Prop({ default: defaultCanvasModel }) value: CanvasModel;
  @Prop({ default: false }) disabled: boolean;
  @Prop({ default: '#ff0000' }) color: string;
  @Prop({ default: 5 }) size: number; // natural numbers
  @Prop({ default: 1 }) opacity: number; // 0-1
  @Prop({ default: 0 }) offsetTop: number;
  @Prop({ default: 0 }) offsetLeft: number;
  @Prop({ default: false }) useStyleBackground: boolean; // enable this if image on default blinking/error
  @Prop({ default: '' }) canvasStyle: string; // custom style for canvas element
  private paint = false;
  private viewInitiated = false;
  width: number = 400;
  height: number = 400;
  vectors = []; // [x, y, drag, size, color, opacity]
  imgLoaded = false;
  imgUrl: string;
  img: HTMLImageElement;
  paper: HTMLCanvasElement;
  ctx: CanvasRenderingContext2D;

  style = '';

  mounted() {
    this.paper = this.$refs.paper as HTMLCanvasElement;
    this.ctx = this.paper.getContext('2d');
    if (!this.ctx) {
      throw new Error('browser not supporting canvas');
    }
    this.paper.ontouchstart = (e) => this.onTouchStart(e);
    this.paper.onmousedown = (e) => this.onMouseDown(e);
    this.paper.onmousemove = (e) => this.onMouseMove(e);
    this.paper.ontouchmove = (e) => this.onTouchMove(e);
    this.paper.ontouchend = (e) => this.disablePaint(e);
    this.paper.onmouseup = (e) => this.disablePaint(e);
    this.paper.onmouseleave = (e) => this.disablePaint(e);
    // set view as loaded & start to draw things
    this.viewInitiated = true;
    this.onValueChange(this.value);
  }

  @Watch('value')
  onValueChange(value: CanvasModel) {
    if (!value) {
      return null;
    }
    if (this.width !== value.width) {
      this.width = value.width;
      this.paper.width = this.width;
    }
    if (this.height !== value.height) {
      this.height = value.height;
      this.paper.height = this.height;
    }
    this.vectors = value.vectors;
    this.imgUrl = value.imgUrl;
    if (this.img?.src !== value.imgUrl) {
      this.loadImage();
    }
    this.style += this.canvasStyle;
    this._reDraw();
  }

  loadImage() {
    if (this.useStyleBackground) {
      this.style = `
      background-image: url('${this.imgUrl}'); 
      background-size: cover; 
      background-repeat: no-repeat;`
      ;      
    } else {
      if (!this.viewInitiated) {
        return null;
      }
      if (!this.imgUrl?.length) {
        return null;
      }
      const w = this.width;
      const h = this.height;
      this.imgLoaded = false;
      this.img = new Image();
      this.img.onload = () => {
        this.imgLoaded = true;
        this.ctx.drawImage(this.img, 0, 0, w, h);
      };
      this.img.src = this.imgUrl;
    }
  }

  onMouseDown(e: MouseEvent) {
    if (this.disabled) {
      return null;
    }
    e.preventDefault();
    const x = e.clientX - this.paper.getBoundingClientRect().x + this.offsetLeft;
    const y = e.clientY - this.paper.getBoundingClientRect().y + this.offsetTop;
    this.paint = true;
    this.addPoint(x, y);
  }

  onTouchStart(e: TouchEvent) {
    if (this.disabled) {
      return null;
    }
    e.preventDefault();
    const x = e.touches[0].clientX - this.paper.getBoundingClientRect().x + this.offsetLeft;
    const y = e.touches[0].clientY - this.paper.getBoundingClientRect().y + this.offsetTop;
    this.paint = true;
    this.addPoint(x, y);
  }

  onMouseMove(e: MouseEvent) {
    if (this.disabled) {
      return null;
    }
    e.preventDefault();
    if (!this.paint) {
      return null;
    }
    const x = e.clientX - this.paper.getBoundingClientRect().x + this.offsetLeft;
    const y = e.clientY - this.paper.getBoundingClientRect().y + this.offsetTop;
    this.addPoint(x, y, true);
  }

  onTouchMove(e: TouchEvent) {
    if (this.disabled) {
      return null;
    }
    e.preventDefault();
    if (!this.paint) {
      return null;
    }
    const x = e.touches[0].clientX - this.paper.getBoundingClientRect().x + this.offsetLeft;
    const y = e.touches[0].clientY - this.paper.getBoundingClientRect().y + this.offsetTop;
    this.addPoint(x, y, true);
  }

  disablePaint(e: MouseEvent | TouchEvent) {
    e.preventDefault();
    this.paint = false;
    this.$emit('input', {
      width: this.width,
      height: this.height,
      vectors: [...this.vectors],
      imgUrl: this.imgUrl,
    });
  }

  addPoint(x: number, y: number, drag: boolean = false) {
    const size = this.size;
    const color = this.color;
    const opacity = this.opacity;
    this.vectors.push([x, y, drag, size, hexToRgb(color), opacity]);
    this._reDraw();
  }

  private _reDraw() {
    if (!this.viewInitiated) {
      return null;
    }
    const w = this.width;
    const h = this.height;
    this.ctx.clearRect(0, 0, w, h);
    if (this.imgLoaded) {
      this.ctx.drawImage(this.img, 0, 0, w, h);
    }
    this.ctx.lineJoin = 'round';

    for (let i = 0; i < this.vectors.length; i++) {
      this.ctx.beginPath();
      if (this.vectors[i][VECTOR_DRAG] && i > 0) {
        this.ctx.moveTo(
          this.vectors[i - 1][VECTOR_X],
          this.vectors[i - 1][VECTOR_Y]
        );
      } else {
        this.ctx.moveTo(
          this.vectors[i][VECTOR_X] - 1,
          this.vectors[i][VECTOR_Y]
        );
      }
      this.ctx.lineTo(this.vectors[i][VECTOR_X], this.vectors[i][VECTOR_Y]);
      this.ctx.closePath();
      const { r, g, b } = this.vectors[i][VECTOR_COLOR];
      this.ctx.strokeStyle = `rgba(${r}, ${g}, ${b}, ${this.vectors[i][VECTOR_OPACITY]})`;
      this.ctx.lineWidth = this.vectors[i][VECTOR_SIZE];
      this.ctx.stroke();
    }
  }
}

// RGB Color format
interface RGB {
  r: number;
  b: number;
  g: number;
}

// hexToRgb will convert hex to RGB color format
function hexToRgb(hex: string): RGB {
  const result = /^#?([a-f\d]{2})([a-f\d]{2})([a-f\d]{2})$/i.exec(hex);
  return result
    ? {
        r: parseInt(result[1], 16),
        g: parseInt(result[2], 16),
        b: parseInt(result[3], 16),
      }
    : null;
}
